##-*- Mode: CPerl -*-

## File: Pd/Patch/Color.pm
## Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
## Description:
##  + Color utilties for Pd::Patch
##======================================================================

package Pd::Patch::Color;
use Exporter;
use IO::File;

##======================================================================
## Exports

our @ISA = qw(Exporter);
our @EXPORT = qw();
our %EXPORT_TAGS =
  (
   color => [
	     qw(rgb2color color2rgb),
	     qw(hex2color color2hex),
	     qw(hex2rgb rgb2hex),
	     qw(pdcolor),
	    ],
   xcolors => [qw(parse_xcolors %xcolors %xcolors_hex)],
  );
our @EXPORT_OK = (map {@$_} values(%EXPORT_TAGS));
$EXPORT_TAGS{all} = [@EXPORT_OK];

##======================================================================
## Default color presets

our %PRESETS_byname =
  (
   white        =>   -262144,  ##-- preset  0: cfcfcf: white
   med_grey     =>   -166441,  ##-- preset  1: 0a0a0a: med_grey
   dark_grey    =>    -66577,  ##-- preset  2: 040404: dark_grey
   pink         =>   -261689,  ##-- preset  3: cf0e0e: pink
   orange_brown =>   -261681,  ##-- preset  4: cf0e0c: orange_brown
   light_yellow =>   -262131,  ##-- preset  5: cfcf8c: light_yellow
   light_green  =>   -225271,  ##-- preset  6: 8dcf8d: light_green
   light_blue   =>   -225280,  ##-- preset  7: 8dcfcf: light_blue
   light_purple =>   -228992,  ##-- preset  8: cd4ecf: light_purple
   lavender     =>   -257472,  ##-- preset  9: 8f8dcf: lavender
   light_grey   =>   -233017,  ##-- preset 10: 0e0e0e: light_grey
   med_grey2    =>   -128992,  ##-- preset 11: c7c7c7: med_grey2
   almost_black =>    -33289,  ##-- preset 12: 020202: almost_black
   red          =>   -258699,  ##-- preset 13: cf8282: red
   orange       =>   -260818,  ##-- preset 14: cfca44: orange
   yellow       =>   -241291,  ##-- preset 15: 8e8e82: yellow
   green        =>    -24198,  ##-- preset 16: 418e41: green
   teal         =>    -44926,  ##-- preset 17: 824f4f: teal
   blue         =>    -62784,  ##-- preset 18: c305cf: blue
   purple       =>   -250685,  ##-- preset 19: 4f030f: purple
   light_gey2   =>   -195568,  ##-- preset 20: cbcbcb: light_gey2
   med_grey3    =>    -99865,  ##-- preset 21: 060606: med_grey3
   black        =>        -1,  ##-- preset 22: 000000: black
   dark_red     =>   -143491,  ##-- preset 23: c88080: dark_red
   brown        =>    -90881,  ##-- preset 24: 850300: brown
   red_brown    =>   -123526,  ##-- preset 25: 878241: red_brown
   dark_green   =>    -42246,  ##-- preset 26: 820541: dark_green
   dark_teal    =>     -1109,  ##-- preset 27: 004405: dark_teal
   dark_blue    =>      -355,  ##-- preset 28: 004188: dark_blue
   dark_purple  =>    -90133,  ##-- preset 29: 850005: dark_purple
  );

our @PRESETS_bynumber =
  (
   -262144,  ##-- preset  0: cfcfcf: white
   -166441,  ##-- preset  1: 0a0a0a: med_grey
    -66577,  ##-- preset  2: 040404: dark_grey
   -261689,  ##-- preset  3: cf0e0e: pink
   -261681,  ##-- preset  4: cf0e0c: orange_brown
   -262131,  ##-- preset  5: cfcf8c: light_yellow
   -225271,  ##-- preset  6: 8dcf8d: light_green
   -225280,  ##-- preset  7: 8dcfcf: light_blue
   -228992,  ##-- preset  8: cd4ecf: light_purple
   -257472,  ##-- preset  9: 8f8dcf: lavender
   -233017,  ##-- preset 10: 0e0e0e: light_grey
   -128992,  ##-- preset 11: c7c7c7: med_grey2
    -33289,  ##-- preset 12: 020202: almost_black
   -258699,  ##-- preset 13: cf8282: read
   -260818,  ##-- preset 14: cfca44: orange
   -241291,  ##-- preset 15: 8e8e82: yellow
    -24198,  ##-- preset 16: 418e41: green
    -44926,  ##-- preset 17: 824f4f: teal
    -62784,  ##-- preset 18: c305cf: blue
   -250685,  ##-- preset 19: 4f030f: purple
   -195568,  ##-- preset 20: cbcbcb: light_gey2
    -99865,  ##-- preset 21: 060606: med_grey3
        -1,  ##-- preset 22: 000000: black
   -143491,  ##-- preset 23: c88080: dark_red
    -90881,  ##-- preset 24: 850300: brown
   -123526,  ##-- preset 25: 878241: red_brown
    -42246,  ##-- preset 26: 820541: dark_green
     -1109,  ##-- preset 27: 004405: dark_teal
      -355,  ##-- preset 28: 004188: dark_blue
    -90133,  ##-- preset 29: 850005: dark_purple
  );

##======================================================================
## Utilities: xcolors

## %xcolors
##  ( "$name" => $pdcolor )
our (%xcolors);

## %xcolors_hex
our (%xcolors_hex);

## $bool = parse_xcolors()
## $bool = parse_xcolors($rgb_txt_file_or_fh)
##  + default searches for 'rgb.txt' in:
##    /usr/X11/lib/X11/rgb.txt
##    /usr/X11R6/lib/X11/rgb.txt
##    /etc/X11/rgb.txt
##  + no implicit color-bashing (with pdcolor()) is performed
sub parse_xcolors {
  my $file = shift;
  if (!defined($file)) {
    foreach my $path ('/usr/X11/lib/X11/rgb.txt',
		      '/usr/X11R6/lib/X11/rb.txt',
		      '/etc/X11/rgb.txt')
      {
	if (-r $path) {
	  $file = $path;
	  last;
	}
      }
    if (!defined($file)) {
      warn(__PACKAGE__, "::parse_xcolors: could not find any rgb.txt file!");
      return undef;
    }
  }
  my $fh = ref($file) ? $file : IO::File->new("<$file");
  if (!defined($fh)) {
    warn(__PACKAGE__, "::parse_xcolors: open failed for '$file': $!");
    return undef;
  }
  my ($line,$r,$g,$b,$name,$pdcolor);
  while (defined($line=<$fh>)) {
    chomp($line);
    $line =~ s/\s*\!.*$//;
    next if ($line =~ /^\s*$/);
    $line =~ s/^\s+//;
    ($r,$g,$b,$name) = split(/\s+/,$line,4);
    $xcolors_hex{$name} = rgb2hex($r,$g,$b);
    $xcolors{$name} = rgb2color($r,$g,$b);
    #$xcolors{$name} = pdcolor(rgb2color($r,$g,$b));
  }
  $fh->close() if (!ref($file));
  return 1;
}

##======================================================================
## Utilities: color

##--------------------------------------------------------------
## Color Format Conversion


## $pd_colorcode = rgb2color($red,$green,$blue)
##  + returns full 24-bit color-code: looks like pd will bash this down to 18 bits
sub rgb2color {
  my ($r,$g,$b) = @_;
  #return -1 * (unpack('V', pack('C4',$r,$g,$b,0)) + 1);
  #return -1 * ( (($r&255)<<16) | (($g&255)<<8) | ($b&255) ) -1;
  return (($r&255)*-65536) + (($g&255)*-256) + (($b&255)*-1) -1;
}

## $pd_safe_colorcode = pdcolor($colorcode_or_preset)
##  + bashes color-code down to 18 bits precision
##    - goofiness: for rgb=(255,255,255) [hex=(0xffffff)]
##      ~ computed color code is   : -16777216 -> 11111111111111111111111100000000 [24 1s]
##      ~ pd reads it ok, but saves:   -262144 -> 11111111111111111100000000000000 [18 1s]
##      ~ looks like we lose 6 bits of precision (probably 2 least significant per color)
##    - more goofiness with AliceBlue: rgb=(240,248,255) [hex=(0xf0f8ff)]
##      ~ computed color code is:    -15792384 -> 
sub pdcolor {
  my $c = shift;
  if ($c >= 0) {
    if ($c > $#PRESETS_bynumber) {
      warn(__PACKAGE__, "::pdcolor(): unknown preset '$c' - using black");
      return -1;
    }
    return $PRESETS_bynumber[$c];
  }
  #my ($r,$g,$b) = color2rgb($c); ##-- no, implicitly bashes (again!)
  ##--
  my $c1 = -1*($c+1);
  my ($r,$g,$b) = ( ($c1/65536)&255, ($c1/256)&255, ($c1&255) );
  #my ($r,$g,$b) = ( ($c>>16)&255, ($c>>8)&255, ($c&255) );

  #return ( (($r&252)*-4096) + (($g&252)*-64) + ($b&252)*-1 )>>2
  return  ( (($r&252)*-1024) + (($g&252)*-16) + (($b&252)/-4) -1);
}


## ($r,$g,$b) = color2rgb($color); ##-- list context
## [$r,$g,$b] = color2rgb($color); ##-- scalar context
sub color2rgb {
  my $color = shift;
  if ($color > 0) {
    if ($color > $#PRESETS_bynumber) {
      warn(__PACKAGE__, "::color2rgb(): unknown preset '$color' - using black");
      return wantarray ? (0,0,0) : [0,0,0];
    }
    $color = $PRESETS_bynumber[$color];
  }
  ##-- true 24-bit version:
  #my $c = -1*($color+1);
  #my @rgb = ( ($c/65536)&255, ($c/256)&255, ($c&255) );
  #my @rgb = ( ($c>>16)&255, ($c>>8)&255, ($c&255) );
  ##-- pd-safe version
  my $c = -1*($color+1);
  my @rgb = ( (($c>>12)&63)<<2, (($c>>6)&63)<<2, ($c&63)<<2 );

  return wantarray ? @rgb : [@rgb];
}

## $pd_colorcode = hex2color($hex_str)
sub hex2color { return -1 * hex($_[0]) -1; }

## ($r,$g,$b) = hex2rgb($hex_str); ##-- list context
## [$r,$g,$b] = hex2rgb($hex_str); ##-- scalar context
#sub hex2rgb { return color2rgb(hex2color($_[0])); }
#sub hex2rgb { return unpack('C3',pack('h6',$_[0])); } ##-- wrong
sub hex2rgb { return unpack('C3',pack('H6',$_[0])); } ##-- wrong

## $hex_str = rgb2hex($r,$g,$b)
#sub rgb2hex { return sprintf("%x%x%x", @_); }
#sub rgb2hex { return unpack('h6', pack('C3', @_)); } ##-- WRONG
sub rgb2hex { return unpack('H6', pack('C3', @_)); }

## $hex_str = color2hex($color);
sub color2hex { return rgb2hex(color2rgb($_[0])); }


1; ##-- be happy

__END__

##======================================================================
## Docs
=pod

=head1 NAME

Pd::Patch - Perl parser and wrapper classes for Pd abstractions

=head1 SYNOPSIS

 use Pd::Patch;

 #... stuff happens...

=cut

##======================================================================
## Description
=pod

=head1 DESCRIPTION

not yet written

=cut

##======================================================================
## Footer
=pod

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

Pd by Miller Puckette and others.

=head1 AUTHOR

Bryan Jurish E<lt>moocow@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT

Copyright (c) 2007, Bryan Jurish.  All rights reserved.

This package is free software.  You may redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1),
pd(1).

=cut
